
import java.awt.*;
import java.awt.event.*;

import net.java.games.jogl.*;

import jahuwaldt.gl.VirtualSphere;
import jahuwaldt.gl.Matrix;


/**
 * VirtualSphereTest.java <BR>
 * original author: Brian Paul (converted to Java by Ron Cemer and Sven Goethel) <P>
 *
 * This is a modification of the standard gears demo so that it
 * uses the Virutal Sphere for converting mouse input to 3D
 * rotation.    --- Joseph A. Huwaldt
 *
 *  @version November 28, 2004
 */
public class VirtualSphereTest {
  public static void main(String[] args) {
    Frame frame = new Frame("Gears with Virtual Sphere");
    GLCanvas canvas = GLDrawableFactory.getFactory().createGLCanvas(new GLCapabilities());

    // Use debug pipeline
    //    canvas.setGL(new DebugGL(canvas.getGL()));
    System.err.println("CANVAS GL IS: " + canvas.getGL().getClass().getName());
    System.err.println("CANVAS GLU IS: " + canvas.getGLU().getClass().getName());

    canvas.addGLEventListener(new GearRenderer());
    frame.add(canvas);
    frame.setSize(400, 400);
    final Animator animator = new Animator(canvas);
    frame.addWindowListener(new WindowAdapter() {
        public void windowClosing(WindowEvent e) {
          animator.stop();
          System.exit(0);
        }
      });
    frame.show();
    animator.start();
  }

  static class GearRenderer implements GLEventListener, MouseListener, MouseMotionListener {
    private int gear1, gear2, gear3;
    private float angle = 0.0f;

	private Point prevMouse;
    private boolean mouseRButtonDown = false;

    // Class variables required to implement the virtual sphere.
    private VirtualSphere vs = new VirtualSphere();
	private Point cueCenter = new Point();
	private int cueRadius;
	private boolean mouseDown = false;
	private float rot_matrix[] = Matrix.identity();

    public void init(GLDrawable drawable) {
	
      //  Initialize the Virtual Sphere.
      Dimension size = drawable.getSize();
      setupVS(size.width, size.height);

      //  Set the initial view angles.
      Matrix.rotateY(Matrix.deg2Rad(30), rot_matrix);
      Matrix.rotateX(Matrix.deg2Rad(20), rot_matrix);

      GL gl = drawable.getGL();
      System.err.println("INIT GL IS: " + gl.getClass().getName());

      float pos[] = { 5.0f, 5.0f, 10.0f, 0.0f };
      float red[] = { 0.8f, 0.1f, 0.0f, 1.0f };
      float green[] = { 0.0f, 0.8f, 0.2f, 1.0f };
      float blue[] = { 0.2f, 0.2f, 1.0f, 1.0f };

      gl.glLightfv(GL.GL_LIGHT0, GL.GL_POSITION, pos);
      gl.glEnable(GL.GL_CULL_FACE);
      gl.glEnable(GL.GL_LIGHTING);
      gl.glEnable(GL.GL_LIGHT0);
      gl.glEnable(GL.GL_DEPTH_TEST);
            
      /* make the gears */
      gear1 = gl.glGenLists(1);
      gl.glNewList(gear1, GL.GL_COMPILE);
      gl.glMaterialfv(GL.GL_FRONT, GL.GL_AMBIENT_AND_DIFFUSE, red);
      gear(gl, 1.0f, 4.0f, 1.0f, 20, 0.7f);
      gl.glEndList();
            
      gear2 = gl.glGenLists(1);
      gl.glNewList(gear2, GL.GL_COMPILE);
      gl.glMaterialfv(GL.GL_FRONT, GL.GL_AMBIENT_AND_DIFFUSE, green);
      gear(gl, 0.5f, 2.0f, 2.0f, 10, 0.7f);
      gl.glEndList();
            
      gear3 = gl.glGenLists(1);
      gl.glNewList(gear3, GL.GL_COMPILE);
      gl.glMaterialfv(GL.GL_FRONT, GL.GL_AMBIENT_AND_DIFFUSE, blue);
      gear(gl, 1.3f, 2.0f, 0.5f, 10, 0.7f);
      gl.glEndList();
            
      gl.glEnable(GL.GL_NORMALIZE);
                
      drawable.addMouseListener(this);
      drawable.addMouseMotionListener(this);
    }
    
    public void reshape(GLDrawable drawable, int x, int y, int width, int height) {
      float h = (float)height / (float)width;
	  
	  GL gl = drawable.getGL();
	  
      gl.glMatrixMode(GL.GL_PROJECTION);

      System.err.println("GL_VENDOR: " + gl.glGetString(GL.GL_VENDOR));
      System.err.println("GL_RENDERER: " + gl.glGetString(GL.GL_RENDERER));
      System.err.println("GL_VERSION: " + gl.glGetString(GL.GL_VERSION));
      System.err.println();
      System.err.println("glLoadTransposeMatrixfARB() supported: " +
                         gl.isFunctionAvailable("glLoadTransposeMatrixfARB"));
      if (!gl.isFunctionAvailable("glLoadTransposeMatrixfARB")) {
        // --- not using extensions
        gl.glLoadIdentity();
      } else {
        // --- using extensions
        final float[] identityTranspose = new float[] {
          1, 0, 0, 0,
          0, 1, 0, 0,
          0, 0, 1, 0,
          0, 0, 0, 1
        };
        gl.glLoadTransposeMatrixfARB(identityTranspose);
      }
      gl.glFrustum(-1.0f, 1.0f, -h, h, 5.0f, 60.0f);
      gl.glMatrixMode(GL.GL_MODELVIEW);
      gl.glLoadIdentity();
      gl.glTranslatef(0.0f, 0.0f, -40.0f);

		// Resize the virtual sphere too!
		setupVS(width, height);
    }

    public void display(GLDrawable drawable) {
      angle += 0.1f;

      GL gl = drawable.getGL();
	  
      gl.glClear(GL.GL_COLOR_BUFFER_BIT | GL.GL_DEPTH_BUFFER_BIT);
            
      gl.glPushMatrix();
      gl.glMultMatrixf(rot_matrix);

      // Render the gears.
      gl.glPushMatrix();
      gl.glTranslatef(-3.0f, -2.0f, 0.0f);
      gl.glRotatef(angle, 0.0f, 0.0f, 1.0f);
      gl.glCallList(gear1);
      gl.glPopMatrix();
            
      gl.glPushMatrix();
      gl.glTranslatef(3.1f, -2.0f, 0.0f);
      gl.glRotatef(-2.0f * angle - 9.0f, 0.0f, 0.0f, 1.0f);
      gl.glCallList(gear2);
      gl.glPopMatrix();
            
      gl.glPushMatrix();
      gl.glTranslatef(-3.1f, 4.2f, 0.0f);
      gl.glRotatef(-2.0f * angle - 25.0f, 0.0f, 0.0f, 1.0f);
      gl.glCallList(gear3);
      gl.glPopMatrix();
            
      gl.glPopMatrix();

      // If the mouse is down, draw the virtual sphere on-screen representation.
      if (mouseDown)	renderVS(drawable);
	  
    }

	/**
	*  Set up the virtual sphere cue center and radius.
	**/
	private void setupVS(int w, int h) {
		cueCenter.x = w/2;
		cueCenter.y = h/2;
		cueRadius = Math.min(w-20, h-20)/2;
	}
        
	/**
	*  Method that sets the drawing characteristics and then
	*  renders the virtual sphere feedback graphics.
	**/
	private void renderVS(GLDrawable drawable) {
	
		GL gl = drawable.getGL();
		
		//	First, save off some state variables that we are going to change.
		boolean lighting = gl.glIsEnabled(gl.GL_LIGHTING);
		boolean texture2D = gl.glIsEnabled(gl.GL_TEXTURE_2D);
		boolean fog = gl.glIsEnabled(gl.GL_FOG);
		
		//	Make sure the GL settings are correct.
		gl.glDisable(gl.GL_LIGHTING);
		gl.glDisable(gl.GL_TEXTURE_2D);
		gl.glDisable(gl.GL_FOG);

		//	Set the line color.
		gl.glColor3f(0.5F, 0.5F, 0.5F);
		
		//	Render the virtual sphere.
		Dimension size = drawable.getSize();
		vs.draw(gl, drawable.getGLU(), size.width, size.height, prevMouse, cueCenter, cueRadius);

		//	Reset state machine settings changed above.
		if (lighting)	gl.glEnable(gl.GL_LIGHTING);
		if (texture2D)	gl.glEnable(gl.GL_TEXTURE_2D);
		if (fog)		gl.glEnable(gl.GL_FOG);
		
	}
		
    public void displayChanged(GLDrawable drawable, boolean modeChanged, boolean deviceChanged) {}

    private void gear(GL gl, float inner_radius,
                      float outer_radius,
                      float width,
                      int teeth,
                      float tooth_depth)
    {
      int i;
      float r0, r1, r2;
      float angle, da;
      float u, v, len;

      r0 = inner_radius;
      r1 = outer_radius - tooth_depth / 2.0f;
      r2 = outer_radius + tooth_depth / 2.0f;
            
      da = 2.0f * (float) Math.PI / teeth / 4.0f;
            
      gl.glShadeModel(GL.GL_FLAT);

      gl.glNormal3f(0.0f, 0.0f, 1.0f);

      /* draw front face */
      gl.glBegin(GL.GL_QUAD_STRIP);
      for (i = 0; i <= teeth; i++)
        {
          angle = i * 2.0f * (float) Math.PI / teeth;
          gl.glVertex3f(r0 * (float)Math.cos(angle), r0 * (float)Math.sin(angle), width * 0.5f);
          gl.glVertex3f(r1 * (float)Math.cos(angle), r1 * (float)Math.sin(angle), width * 0.5f);
          if(i < teeth)
            {
              gl.glVertex3f(r0 * (float)Math.cos(angle), r0 * (float)Math.sin(angle), width * 0.5f);
              gl.glVertex3f(r1 * (float)Math.cos(angle + 3.0f * da), r1 * (float)Math.sin(angle + 3.0f * da), width * 0.5f);
            }
        }
      gl.glEnd();

      /* draw front sides of teeth */
      gl.glBegin(GL.GL_QUADS);
      for (i = 0; i < teeth; i++)
        {
          angle = i * 2.0f * (float) Math.PI / teeth;
          gl.glVertex3f(r1 * (float)Math.cos(angle), r1 * (float)Math.sin(angle), width * 0.5f);
          gl.glVertex3f(r2 * (float)Math.cos(angle + da), r2 * (float)Math.sin(angle + da), width * 0.5f);
          gl.glVertex3f(r2 * (float)Math.cos(angle + 2.0f * da), r2 * (float)Math.sin(angle + 2.0f * da), width * 0.5f);
          gl.glVertex3f(r1 * (float)Math.cos(angle + 3.0f * da), r1 * (float)Math.sin(angle + 3.0f * da), width * 0.5f);
        }
      gl.glEnd();
    
      /* draw back face */
      gl.glBegin(GL.GL_QUAD_STRIP);
      for (i = 0; i <= teeth; i++)
        {
          angle = i * 2.0f * (float) Math.PI / teeth;
          gl.glVertex3f(r1 * (float)Math.cos(angle), r1 * (float)Math.sin(angle), -width * 0.5f);
          gl.glVertex3f(r0 * (float)Math.cos(angle), r0 * (float)Math.sin(angle), -width * 0.5f);
          gl.glVertex3f(r1 * (float)Math.cos(angle + 3 * da), r1 * (float)Math.sin(angle + 3 * da), -width * 0.5f);
          gl.glVertex3f(r0 * (float)Math.cos(angle), r0 * (float)Math.sin(angle), -width * 0.5f);
        }
      gl.glEnd();
    
      /* draw back sides of teeth */
      gl.glBegin(GL.GL_QUADS);
      for (i = 0; i < teeth; i++)
        {
          angle = i * 2.0f * (float) Math.PI / teeth;
          gl.glVertex3f(r1 * (float)Math.cos(angle + 3 * da), r1 * (float)Math.sin(angle + 3 * da), -width * 0.5f);
          gl.glVertex3f(r2 * (float)Math.cos(angle + 2 * da), r2 * (float)Math.sin(angle + 2 * da), -width * 0.5f);
          gl.glVertex3f(r2 * (float)Math.cos(angle + da), r2 * (float)Math.sin(angle + da), -width * 0.5f);
          gl.glVertex3f(r1 * (float)Math.cos(angle), r1 * (float)Math.sin(angle), -width * 0.5f);
        }
      gl.glEnd();
    
      /* draw outward faces of teeth */
      gl.glBegin(GL.GL_QUAD_STRIP);
      for (i = 0; i < teeth; i++)
        {
          angle = i * 2.0f * (float) Math.PI / teeth;
          gl.glVertex3f(r1 * (float)Math.cos(angle), r1 * (float)Math.sin(angle), width * 0.5f);
          gl.glVertex3f(r1 * (float)Math.cos(angle), r1 * (float)Math.sin(angle), -width * 0.5f);
          u = r2 * (float)Math.cos(angle + da) - r1 * (float)Math.cos(angle);
          v = r2 * (float)Math.sin(angle + da) - r1 * (float)Math.sin(angle);
          len = (float)Math.sqrt(u * u + v * v);
          u /= len;
          v /= len;
          gl.glNormal3f(v, -u, 0.0f);
          gl.glVertex3f(r2 * (float)Math.cos(angle + da), r2 * (float)Math.sin(angle + da), width * 0.5f);
          gl.glVertex3f(r2 * (float)Math.cos(angle + da), r2 * (float)Math.sin(angle + da), -width * 0.5f);
          gl.glNormal3f((float)Math.cos(angle), (float)Math.sin(angle), 0.0f);
          gl.glVertex3f(r2 * (float)Math.cos(angle + 2 * da), r2 * (float)Math.sin(angle + 2 * da), width * 0.5f);
          gl.glVertex3f(r2 * (float)Math.cos(angle + 2 * da), r2 * (float)Math.sin(angle + 2 * da), -width * 0.5f);
          u = r1 * (float)Math.cos(angle + 3 * da) - r2 * (float)Math.cos(angle + 2 * da);
          v = r1 * (float)Math.sin(angle + 3 * da) - r2 * (float)Math.sin(angle + 2 * da);
          gl.glNormal3f(v, -u, 0.0f);
          gl.glVertex3f(r1 * (float)Math.cos(angle + 3 * da), r1 * (float)Math.sin(angle + 3 * da), width * 0.5f);
          gl.glVertex3f(r1 * (float)Math.cos(angle + 3 * da), r1 * (float)Math.sin(angle + 3 * da), -width * 0.5f);
          gl.glNormal3f((float)Math.cos(angle), (float)Math.sin(angle), 0.0f);
        }
      gl.glVertex3f(r1 * (float)Math.cos(0), r1 * (float)Math.sin(0), width * 0.5f);
      gl.glVertex3f(r1 * (float)Math.cos(0), r1 * (float)Math.sin(0), -width * 0.5f);
      gl.glEnd();
    
      gl.glShadeModel(GL.GL_SMOOTH);
    
      /* draw inside radius cylinder */
      gl.glBegin(GL.GL_QUAD_STRIP);
      for (i = 0; i <= teeth; i++)
        {
          angle = i * 2.0f * (float) Math.PI / teeth;
          gl.glNormal3f(-(float)Math.cos(angle), -(float)Math.sin(angle), 0.0f);
          gl.glVertex3f(r0 * (float)Math.cos(angle), r0 * (float)Math.sin(angle), -width * 0.5f);
          gl.glVertex3f(r0 * (float)Math.cos(angle), r0 * (float)Math.sin(angle), width * 0.5f);
        }
      gl.glEnd();
    }

    // Methods required for the implementation of MouseListener
    public void mouseEntered(MouseEvent e) {}
    public void mouseExited(MouseEvent e) {}

    public void mousePressed(MouseEvent e) {
      mouseDown = true;
      prevMouse = e.getPoint();
      if ((e.getModifiers() & e.BUTTON3_MASK) != 0) {
        mouseRButtonDown = true;
      }
    }
    
    public void mouseReleased(MouseEvent e) {
      mouseDown = false;
      if ((e.getModifiers() & e.BUTTON3_MASK) != 0) {
        mouseRButtonDown = false;
      }
    }
    
    public void mouseClicked(MouseEvent e) {}
    
	// The incremental rot-matrix calculated by the virtual sphere.
	private float[] mouseMtx = new float[16];

    // Methods required for the implementation of MouseMotionListener
    public void mouseDragged(MouseEvent e) {
		Point newMouse = e.getPoint();

		if (newMouse.x != prevMouse.x || newMouse.y != prevMouse.y) {
		
			// Use virtual sphere to determine an incremental rotation matrix.
			vs.makeRotationMtx(prevMouse, newMouse, cueCenter, cueRadius, mouseMtx);
		
			// Multiply overall rotation matrix by incremental rot-matrix to get new rotation.
			rot_matrix = Matrix.multiply(rot_matrix, mouseMtx);
			fixRotationMatrix();
				
			prevMouse = newMouse;
	   }
    }
    
	private void fixRotationMatrix() {
		// Fix any problems with the rotation matrix.
		rot_matrix[3] =
		rot_matrix[7] =
		rot_matrix[11] =
		rot_matrix[12] =
		rot_matrix[13] =
		rot_matrix[14] = 0.0f;
		rot_matrix[15] = 1.0f;
		float fac;
		if ((fac = (float)Math.sqrt
			((rot_matrix[0]*rot_matrix[0]) +
			 (rot_matrix[4]*rot_matrix[4]) +
			 (rot_matrix[8]*rot_matrix[8]))) != 1.0f)
		{
			if (fac != 0.0f)
			{
				fac = 1.0f/fac;
				rot_matrix[0] *= fac;
				rot_matrix[4] *= fac;
				rot_matrix[8] *= fac;
			}
		}
		if ((fac = (float)Math.sqrt
			((rot_matrix[1]*rot_matrix[1]) +
			 (rot_matrix[5]*rot_matrix[5]) +
			 (rot_matrix[9]*rot_matrix[9]))) != 1.0f)
		{
			if (fac != 0.0f)
			{
				fac = 1.0f/fac;
				rot_matrix[1] *= fac;
				rot_matrix[5] *= fac;
				rot_matrix[9] *= fac;
			}
		}
		if ((fac = (float)Math.sqrt
			((rot_matrix[2]*rot_matrix[2]) +
			 (rot_matrix[6]*rot_matrix[6]) +
			 (rot_matrix[10]*rot_matrix[10]))) != 1.0f)
		{
			if (fac != 0.0f)
			{
				fac = 1.0f/fac;
				rot_matrix[2] *= fac;
				rot_matrix[6] *= fac;
				rot_matrix[10] *= fac;
			}
		}
	}

    public void mouseMoved(MouseEvent e) {}
  }
}

